/*
   Direct3D 9 Matrices Demo.
   Created by the Programming Ace.
   www.UltimateGameProgramming.com

   The purpose of this demo is to use matrices to set the position of the view and
   to use a matrix to rotate the triangle on the screen.  Basically you use matrices
   to change the position and orientation of something.  In this demo we are going to use it to
   change the view (the spot the camera is looking at and positioned) and another matrix
   to rotate the triangle.
*/


// Demo: Direct3D 9 Matrices.


#define WIN32_LEAN_AND_MEAN
#define VC_LEANMEAN

#include<d3dx9.h>


// Our vertex stucture define.
#define D3DFVF_D3DVertex (D3DFVF_XYZ|D3DFVF_DIFFUSE)


// Function Prototypes...
bool InitializeDirect3D(HWND hwnd, bool fullscreen);
bool InitializeObject();
void SetRotation();
void SetCamera();
void SetPerspective();
void RenderScence();
void ShutdownDirect3D();


// Direct3D objects.
LPDIRECT3D9 Direct3D_Object = NULL;
LPDIRECT3DDEVICE9 D3D_Device = NULL;
LPDIRECT3DVERTEXBUFFER9 Vertex_Buffer = NULL;


// This is the rotation angle.
float angle = 0.0f;

// Vertex stucture with xyz position and color.
struct Vertex
{
   FLOAT x, y, z;
   DWORD color;
};


LRESULT CALLBACK WndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    switch(message)
       {
		   case WM_KEYUP:
            // If the user presses the escape key then exit the application.
            if(wParam == VK_ESCAPE)
               PostQuitMessage(0);
            break;

         case WM_DESTROY:
         case WM_CLOSE:
            PostQuitMessage(0);
            break;

            default:
               break;
       }

    return DefWindowProc( hwnd, message, wParam, lParam );
}


int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nShowCmd)
{
   MSG msg;                                           // Message object.
   HWND hwnd;                                         // Handle to the window.
   WNDCLASSEX windowClass;                            // Window class object.
   bool done = false;                                 // True then exit.

   // This is the Window class.
   windowClass.cbSize = sizeof(WNDCLASSEX);           // size of the WNDCLASSEX structure.
   windowClass.style = CS_HREDRAW | CS_VREDRAW;       // style of the window.
   windowClass.lpfnWndProc = WndProc;                 // Address to the windows procedure.
   windowClass.cbClsExtra = 0;                        // Extra class information.
   windowClass.cbWndExtra = 0;                        // Extra window information.
   windowClass.hInstance = hInstance;                 // Handle of application Instance.
   windowClass.hIcon = LoadIcon(NULL, IDI_APPLICATION);// Window Icon.
   windowClass.hCursor = LoadCursor(NULL, IDC_ARROW); // Load mouse cursor.
   windowClass.hbrBackground = NULL;                  // Background color.
   windowClass.lpszMenuName = NULL;                   // Menu.
   windowClass.lpszClassName = "UGPClass";            // Name of the window class.
   windowClass.hIconSm = LoadIcon(NULL, IDI_APPLICATION);// Minimized window icon.

   // You must register you class with Windows.
   if(!RegisterClassEx(&windowClass)) return 0;

   // Create the window.
   hwnd = CreateWindowEx(NULL,                                       // The extended style.
                         "UGPClass",                                 // Window class.
                         "Direct3D Matrices Demo - by The Programming Ace",// Window name.
                          WS_OVERLAPPEDWINDOW | WS_VISIBLE |         // Window style.
                          WS_SYSMENU |WS_CLIPCHILDREN |              // Window style.
                          WS_CLIPSIBLINGS,                           // Window style.
                          100, 100,                                  // X, Y coords.
                          640, 480,                                  // Window size.
                          NULL,                                      // Handle to parent window.
                          NULL,                                      // Menu.
                          hInstance,                                 // Handle to app instance.
                          NULL);                                     // Pointer to window.

   // If there was an error with creating the window, then close the program.
   if(!hwnd) return 0;

   ShowWindow(hwnd, SW_SHOW);    // Show the window.
   UpdateWindow(hwnd);           // Update its display.

   done = false;  // false = run program, true means stop.

   // Initialize Direct3D.  If fails then we don't want to run the application.
   if(!InitializeDirect3D(hwnd, false)) done = true;

   // Application loop.
   while(!done)
      {
         if(PeekMessage(&msg, NULL, NULL, NULL, PM_REMOVE))
            {
               // If a quit message is received then stop rendering and quit the app.
               if(msg.message == WM_QUIT)
                  {
                     done = true;
                  }

               TranslateMessage(&msg);
               DispatchMessage(&msg);
            }
         else
            {
               RenderScence();
            }
      }

    // Here we shutdown Direct 3D.
	ShutdownDirect3D();

   // Here we unregister the window class with the OS.
   UnregisterClass("UGPClass", windowClass.hInstance);

   return (int)msg.wParam;
}


bool InitializeDirect3D(HWND hwnd, bool fullscreen)
{
   // This object will allow us to set the display mode of the screen.
   D3DDISPLAYMODE DisplayMode;

   // This will allow us to set the parameters of the screen.
   D3DPRESENT_PARAMETERS Present_Parameters;

   // This is used to get the capabilities of the hardware.
   D3DCAPS9 D3DCaps;

   // It is always a good idea to clear out memory in object although not necessary.
   ZeroMemory(&Present_Parameters, sizeof(Present_Parameters));

   // Create the Direct3D object to get everything started.
   Direct3D_Object = Direct3DCreate9(D3D_SDK_VERSION);

   // Error checking.  Make sure that it was successful.
   if(Direct3D_Object == NULL)
      {
         MessageBox(NULL, "Error, couldn't initialize DirectX!?!",
                    "Error!", MB_OK);
         return false;
      }

   // This function will get the display mode of the device and place it in DisplayMode.
   if(FAILED(Direct3D_Object->GetAdapterDisplayMode(D3DADAPTER_DEFAULT, &DisplayMode)))
      {
         MessageBox(NULL, "Error setting the display mode.", "Error!", MB_OK);
         return false;
      }

   // Get the capabilities of the hardware.
	if(FAILED(Direct3D_Object->GetDeviceCaps(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, &D3DCaps)))
		{
		   return false;
	   }

   // Test which is supported, hardware or software vertex processing.
	DWORD VertexProcessing = 0;

	if(D3DCaps.VertexProcessingCaps != 0)
		VertexProcessing |= D3DCREATE_HARDWARE_VERTEXPROCESSING;
	else
		VertexProcessing |= D3DCREATE_SOFTWARE_VERTEXPROCESSING;

   // Here we are setting the applications parameters...
   if(fullscreen)
      {
         Present_Parameters.Windowed = FALSE;               // Window mode (fullscreen).
         Present_Parameters.BackBufferWidth = 640;
         Present_Parameters.BackBufferHeight = 480;
      }
   else
      Present_Parameters.Windowed = TRUE;                   // Window mode (not fullscreen).
   Present_Parameters.SwapEffect = D3DSWAPEFFECT_DISCARD;   // Dealing with animation (see doc).
   Present_Parameters.BackBufferFormat = DisplayMode.Format;// Render to the area of the screen.
   Present_Parameters.BackBufferCount = 1;                  // Number of back buffers.
   Present_Parameters.EnableAutoDepthStencil = TRUE;        // Check documentation.
   Present_Parameters.AutoDepthStencilFormat = D3DFMT_D16;  // Check documentation.


   // Now we must create the rendering device.
   if(FAILED(Direct3D_Object->CreateDevice(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, hwnd,
                                           VertexProcessing,
                                           &Present_Parameters, &D3D_Device)))
      {
         MessageBox(NULL, "CreateDevice() failed!  Make sure you have DirectX 9.",
                    "Error!", MB_OK);
         return false;
      }


   // One last check to be sure.
   if(D3D_Device == NULL)
      {
         MessageBox(NULL, "D3D_Device is equal to NULL!?!", "Error!", MB_OK);
         return false;
      }

   // SetRenderState() will set a parameter for the rendering.  Here we turn off lighting.
   D3D_Device->SetRenderState(D3DRS_LIGHTING, FALSE);

   // Disable culling.  This will render the front and back of the triangles.
	D3D_Device->SetRenderState(D3DRS_CULLMODE, D3DCULL_NONE);

   // Here we set up the object this demo will render.
   if(!InitializeObject()) return false;

   // Here we set the demos perpective.  We will use a matrix to determine where
   // the sight that we can see starts and ends.  For ex. how far back can we see (500).
   SetPerspective();

   return true;
}


bool InitializeObject()
{
   // Triangle data for its size and color.  Each point has a x, y, and z position
   // as well as a Red, Green, and Blue color (D3DCOLOR_XRGB(Red, Green, Blue).
	Vertex Triangle_Data[3] =
	{
		{-0.5f, -0.5f, 0.0f, D3DCOLOR_XRGB(255,255,0)},
      {0.5f, -0.5f, 0.0f, D3DCOLOR_XRGB(255,255,0)},
	   {0.0f, 0.3f, 0.0f, D3DCOLOR_XRGB(0,0,255)}
	};

   // Create the vertex buffer that will hold the triangle.
	if(FAILED(D3D_Device->CreateVertexBuffer(sizeof(Triangle_Data),
                                            0, D3DFVF_D3DVertex,
                                            D3DPOOL_DEFAULT, &Vertex_Buffer, NULL)))
      {
         return false;
      }
	
   // Pointer to the buffer.
   Vertex* Vertices;

   // Lock the buffer we can write to it.
   if(FAILED(Vertex_Buffer->Lock(0, sizeof(Triangle_Data), (void**)&Vertices, 0)))
      return false;

   // Here we copy the triangle's data into the vertex buffer.
   memcpy(Vertices, Triangle_Data, sizeof(Triangle_Data));

   // Unlock when your done coping data into the buffer.
   Vertex_Buffer->Unlock();

   return true;
}


void SetRotation()
{
   // This will change the rotation position of the object we are drawing after
   // each frame.  This gives us the animation of the object being rotation along
   // the X axis.

   // This will hold the world matrix.
   D3DXMATRIXA16 World;

   // Set the rotation value to the matrix.
   D3DXMatrixRotationZ(&World, angle);

   // Set the matrix.
   D3D_Device->SetTransform(D3DTS_WORLD, &World);

   // Add 0.1 to angle to rotate the triangle even more the next frame.
   angle += 0.1f;

   // If the angle is 360 degrees or higher we want to reset the angle variable.
   // 360 degrees is a full circle.  After 360 you are back to where you started.
   // For example stand up and turn around.  Once you are done you have turned a full
   // 360 degrees.
   if(angle >= 360) angle = 0.0f;
}


void SetCamera()
{
   // This matrix will set up our view.  Similiar to a camera.  Here we are
   // telling Direct3D where to place the "camera" before drawing.

   // This will hold the view ("camera") matrix.
   D3DXMATRIXA16 View;

   // Eye is the position, LookAt is the point the camera is looking at, and Up
   // just tells Direct3D which way is up (positive 1 and above is...well up).

   // Think of D3DXVECTOR3 as being a class that holds 3 floats.  Here
   // we are basically moving the camera back 5 units.
   D3DXVECTOR3 Eye(0.0f, 0.0f, -5.0f);     // Position of camera.
   D3DXVECTOR3 LookAt(0.0f, 0.0f, 0.0f);   // Point camera is looking at.
   D3DXVECTOR3 Up(0.0f, 1.0f, 0.0f);       // Which way is up.

   // Set all the values this matrix will need to know (position, look at point, etc).
   D3DXMatrixLookAtLH(&View, &Eye, &LookAt, &Up);

   // Move the "camera" to the position and values we set above.
   D3D_Device->SetTransform(D3DTS_VIEW, &View);
}


void SetPerspective()
{
   // This matrix will do nothing more but tell Direct3D how far back it can see
   // and how close it can see.
   D3DXMATRIXA16 Projection;

   // Set the info into the matrix.  Anything past 500.0 along the Z axis will
   // not be seen.  Same goes with anything in front of 0.0.  You can set these
   // numbers to what ever you want.  The first parameter is the matrix, the second
   // is the field of view, next the aspect ratio, then the near plane, and last the
   // far plane.
   D3DXMatrixPerspectiveFovLH(&Projection, 1.0f, 640/480, 0.0f, 500.0f);

   // Set the projection.
   D3D_Device->SetTransform(D3DTS_PROJECTION, &Projection);
}


void RenderScence()
{
   // Clear the screen to black.
   D3D_Device->Clear(0, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER,
                     D3DCOLOR_XRGB(0,0,0), 1.0f, 0);

   // Tell Direct 3D we are now going to start drawing.
   D3D_Device->BeginScene();

      // Set the world matrix so we can rotate and animate our triangle.
      SetRotation();

      // Set the view matrix (camera position).
      SetCamera();

	   // This will bind the vertex data in the buffer to the Direct3D device.
      D3D_Device->SetStreamSource(0, Vertex_Buffer, 0, sizeof(Vertex));

      // Set the vertex stream that was defined at the top.
      D3D_Device->SetFVF(D3DFVF_D3DVertex);

      // This will draw everything in the buffer.
      D3D_Device->DrawPrimitive(D3DPT_TRIANGLELIST, 0, 1);
    
   // We are done drawing for this scene.
   D3D_Device->EndScene();
    
   // Swap the old frame with the new one.
   D3D_Device->Present(NULL, NULL, NULL, NULL);
}


void ShutdownDirect3D()
{
   // Here we release the Direct3D objects.
   if(D3D_Device != NULL)
      {
         D3D_Device->Release();
         D3D_Device = NULL;
      }

   if(Direct3D_Object != NULL)
      {
         Direct3D_Object->Release();
         Direct3D_Object = NULL;
      }

   if(Vertex_Buffer != NULL)
      {
         // Release the buffer.
	      Vertex_Buffer->Release();
         Vertex_Buffer = NULL;
      }
}


// Recap:

// Play with these values in order to get a better idea of what is going on.
// All that this demo is doing is moving the view back 5 (the camera)
// and rotating the triangle.  As long as all objects have their own matrix
// you can rotate some while other stay still or they can have their own rotations.
// After a number of demos I will create a camera demo that will change
// the values in the view matrix so that you could move around a 3D world (space).
// This demo did not need the view matrix part but I added it just to show another
// way matrices are used.


// Copyright September 2004
// All Rights Reserved!
// Allen Sherrod
// ProgrammingAce@UltimateGameProgramming.com
// www.UltimateGameProgramming.com